# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import os
import sys
import shlex
import argparse
from functools import lru_cache

from prompt_toolkit import PromptSession
from prompt_toolkit.formatted_text.html import HTML

from kpd.db import DelayedPyKeePass
from kpd.utils import eprint
from kpd.commands import (
    CommandError, autotype, show, echo, sleep, ls, help_, exit, unlock, lock,
    open_)
from kpd.completion import CommandCompleter
from kpd._version import version


class ArgumentParserError(Exception): pass


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


@lru_cache(maxsize=32)
def ps1(prompt, dbpath):
    if not dbpath:
        dbpath = ''
    home = os.path.expanduser('~')
    if dbpath.startswith(home):
        dbpath = dbpath.replace(home, '~')

    return HTML(prompt.format(dbpath))


def readprompt(kp, prompt, parsers):
    compl = CommandCompleter(kp, (p for p in parsers if p))
    session = PromptSession(completer=compl, complete_while_typing=False)

    def _fn():
        while True:
            try:
                yield session.prompt(ps1(prompt, kp.db))
            except EOFError:
                break
            except KeyboardInterrupt:
                continue
    return _fn


def readinput():
    while True:
        try:
            yield input()
        except EOFError:
            break


def run_shell(kp, prompt, cmd_parser, parsers):
    if sys.stdin.isatty():
        reader = readprompt(kp, prompt, parsers)
    else:
        reader = readinput

    for text in reader():
        cmd = shlex.split(text)
        if not cmd:
            continue

        try:
            cargs = cmd_parser.parse_args(cmd)
            if cargs.func(kp, cargs) is False:
                break
        except ArgumentParserError as e:
            eprint(str(e))
        except CommandError as e:
            eprint(str(e))


def prepare_command_parser():
    cp = ThrowingArgumentParser(prog='', add_help=False)
    sp = cp.add_subparsers(required=True)

    parsers = {}
    parsers[None] = cp

    # helper function which automatically creates help-friendly parsers
    def add_parser(command, *a, **kw):
        kw['add_help'] = False

        descr = kw.get('description')
        if descr:
            kw['description'] = '{}\n\n{}'.format(kw['help'], descr)
        else:
            kw['description'] = kw['help']

        parser = sp.add_parser(command, *a, **kw)
        parsers[command] = parser
        return parser

    ######### open
    open_sp = add_parser('open', help='Change currently opened database.')
    open_sp.add_argument('filepath', help='path to database file.')
    open_sp.set_defaults(func=open_)

    ######### unlock
    unlock_sp = add_parser('unlock', help='Unlock currently opened database.')
    unlock_sp.add_argument('--keyfile', default='',
                           help='key file used for unlocking database')
    unlock_sp.set_defaults(func=unlock)

    ######### lock
    lock_sp = add_parser('lock', help='Lock a database.')
    lock_sp.set_defaults(func=lock)

    ######### ls
    ls_sp = add_parser('ls', help='List contents of database.')
    ls_sp.add_argument('glob', nargs='?', default='*',
                       help='display only entries which match glob expression')
    ls_sp.set_defaults(func=ls)

    ######### show
    show_sp = add_parser('show', help='Show contents of entry.',
        description='Search is case-sensitive.')
    show_sp.add_argument('path', help='path which should be shown')
    show_sp.add_argument('fields', nargs='*',
        help='only display certain fields')
    show_sp.add_argument('-n', '--no-field-name', action='store_true',
        help='hide field name when printing entry fields.')
    show_sp.set_defaults(func=show)

    ######### autotype
    at_sp = add_parser('autotype', help='Auto-type sequence of entry fields.',
        description='This simulates keypresses to any currently open window. '
                    'It\'s particularily useful when kpsh is run from a script '
                    'or keypress in non-interactive mode (`-c` switch). If '
                    '`-s` is given, it will be used as auto-type sequence. '
                    'Otherwise sequence defined for selected entry will be '
                    'used or the default one if there is none (`-d`).')
    at_sp.add_argument('path', help='path of entry to auto-type')
    at_sp.add_argument('-s', '--sequence', help='override auto-type sequence')
    at_sp.add_argument('-d', '--default',
        default='{USERNAME}{TAB}{PASSWORD}{ENTER}',
        help='default auto-type sequence used when entry doesn\'t specify '
             'sequence itself.')
    at_sp.add_argument('-D', '--delay', type=int, default=40,
        help='delay beteen simulated keypresses')
    at_sp.add_argument('-f', '--force', action='store_true',
        help='force auto-type for entries for which auto-type was disabled')
    at_sp.set_defaults(func=autotype)

    ######### exit
    exit_sp = add_parser('exit', help='Exit shell.')
    exit_sp.set_defaults(func=exit)

    ######### echo
    echo_sp = add_parser('echo', help='Display a message.')
    echo_sp.add_argument('message', nargs='*', help='message to be displayed')
    echo_sp.set_defaults(func=echo)

    ######### sleep
    sleep_sp = add_parser('sleep', help='Sleep for a given number of seconds.',
        description='Seconds might be a floating number when fractions of '
                    'second are needed.')
    sleep_sp.add_argument('secs', type=float, help='seconds to sleep')
    sleep_sp.set_defaults(func=sleep)

    ######### help
    help_sp = add_parser('help', help='Show help for any message.')
    help_sp.add_argument('command', nargs='?')
    help_sp.set_defaults(func=lambda *a, parsers=parsers: help_(*a, parsers))

    return cp, parsers


def prepare_args():
    ap = argparse.ArgumentParser(description='KeePass database shell access.')
    ap.add_argument('db', nargs='?', help='path to KeePass database.')

    ap.add_argument('--password', default=None,
        help='Database password.')
    ap.add_argument('--pw-cmd', default=None,
        help='Password will be obtained from the output of this command.')
    ap.add_argument('--keyfile', default=None,
        help='Key file for unlocking database.')
    ap.add_argument('--pinentry',
        help='Command used to run pinentry.')
    ap.add_argument('-c', '--command', action='append',
        help='Command to execute. If command contains spaces, it must be '
             'enclosed in double quotes. kpsh will be started in '
             'non-interactive mode.')
    ap.add_argument('--prompt', default='<style fg="ansiblue">{}</style>> ',
        help='Text used by shell for prompt.')
    ap.add_argument('--version', action='version',
                    version='%(prog)s {}'.format(version))
    return ap.parse_args()

def main():
    args = prepare_args()

    kp = DelayedPyKeePass(args)

    cmd_parser, parsers = prepare_command_parser()
    if args.command:
        for command in args.command:
            command = shlex.split(command)
            cargs = cmd_parser.parse_args(command)
            cargs.func(kp, cargs)
    else:
        run_shell(kp, args.prompt, cmd_parser, parsers)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import os
import sys
import subprocess
import operator
import collections

from prompt_toolkit import prompt

from kpd.utils import eprint

class DatabaseError(Exception):
    pass

# Delays importing of PyKeePass which by itself adds ~150 ms to startup...
class DelayedPyKeePass:
    def __init__(self, args):
        self._password = args.password
        self._keyfile = args.keyfile

        self._pw_cmd = args.pw_cmd
        self._pinentry = args.pinentry

        self._db = None
        self._kp = None
        self._entries  = None

        if args.db:
            self.change_db(args.db)

    def _getpass(self):
        if self._password is not None:
            yield self._password
        if self._pw_cmd is not None:
            yield readpass(self._pw_cmd)
        if self._password is None and self._pw_cmd is None \
                and self._keyfile is not None:
            # password=None supports Composite Master Key with only a keyfile
            # https://keepass.info/help/base/keys.html
            yield None
        yield askpass(self._db, self._pinentry)

    def unlock(self):
        if self._kp:
            return True

        if not self._db:
            eprint('No database is opened.')
            return False

        from pykeepass import PyKeePass

        kp = None
        for pw in self._getpass():
            try:
                kp = PyKeePass(self._db,
                               password=pw,
                               keyfile=self._keyfile)
            except Exception as e:
                pass
            else:
                break

        if kp is None:
            eprint('Failed unlocking database {}.'.format(self._db))
            return False

        self._kp = kp
        self._entries = None
        return True

    def lock(self):
        self._kp = None
        self._entries = None
        self._password = None
        self._keyfile = None

    def change_db(self, db):
        self.lock()
        if not os.path.isfile(db):
            eprint('Invalid database: \'{}\''.format(db))
            return
        self._db = db

    def change_credentials(self, password=None, keyfile=None):
        if not self.locked:
            raise DatabaseError(
                'Database must be locked before changing credentials '
                'used for unlocking.')

        self._password = password
        self._keyfile = keyfile

    @property
    def locked(self):
        return self._kp is None

    @property
    def db(self):
        return self._db

    @property
    def kp(self):
        self.unlock()
        return self._kp

    @property
    def entries(self):
        if not self.kp:
            return collections.OrderedDict()

        if self._entries is None:
            self._entries = collections.OrderedDict(
                (e.path, e)
                for e in sorted(self.kp.entries,
                                key=operator.attrgetter('path')))

        return self._entries

    @property
    def paths(self):
        return [e for e in self.entries.keys()]

    def iter_paths(self):
        for entry in self.entries:
            yield entry

def askpass(db, pinentry):
    if pinentry:
        return _askpass_pinentry(db, pinentry)
    return _askpass_prompt()


def readpass(cmd):
    cp = subprocess.run(cmd, capture_output=True, text=True, shell=True)
    if cp.returncode != 0 or not cp.stdout:
        return None
    return cp.stdout


def _askpass_pinentry(db, pinentry):
    pein = 'setdesc Enter password for {}\ngetpin\n'.format(db)
    cp = subprocess.run(pinentry, input=pein, capture_output=True,
                        text=True)

    if cp.returncode != 0 or not cp.stdout:
        return None

    lines = cp.stdout.splitlines()
    passline = [line for line in lines if line.startswith ('D ')][0]
    return passline.partition(' ')[-1]


def _askpass_prompt():
    if not sys.stdin.isatty():
        return None

    try:
        password = prompt('Password: ', is_password=True)
        return password
    except:
        return None



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import sys
import os
import subprocess
import time
import fnmatch

from kpd.utils import eprint

class CommandError(Exception):
    pass

def tokenize(seq):
    tokens = []

    i = 0
    while i < len(seq):
        start = seq.find('{', i)

        if start != -1:
            if i < start:
                tokens.append(seq[i:start])

            end = seq.find('}', start)
            nend = end + 1

            if end == -1:
                end = len(seq) - 1
            elif end == start + 1 and len(seq) > nend and seq[nend] == '}':  # {}}
                end = nend

            tokens.append(seq[start:end + 1])
            i = end + 1
        else:
            tokens.append(seq[i:])
            i = len(seq)

    return tokens


def xdotool(*args):
    cmd = ['/usr/bin/xdotool']
    cmd.extend(args)
    subprocess.run(cmd)


def autotype(kp, args):
    from kpd.autotype.xdotoolkeys import XDOTOOL_KEYS
    from kpd.autotype.placeholders import replace_placeholder
    from kpd.autotype.commands import run_command

    delay = str(args.delay)
    entry = _get(args.path, kp)

    if not entry.autotype_enabled and not args.force:
        eprint('Autotype disabled for {}. '
               'Use -f to force autotype.'.format(args.path))
        return

    sequence = args.sequence if args.sequence else entry.autotype_sequence
    if not sequence:
        sequence = args.default

    for token in tokenize(sequence):
        if token.startswith('{') and token.endswith('}'):
            if run_command(token):
                continue

            placeholder = replace_placeholder(entry, token)
            if placeholder is not None:
                xdotool('type', '--delay', delay, placeholder)
                continue

            specialkey = XDOTOOL_KEYS.get(token)
            if specialkey is not None:
                xdotool('key', '--clearmodifiers', '--delay', delay, specialkey)
                continue

            eprint('Unsupported keyword: {}'.format(token))
        else:
            xdotool('type', '--delay', delay, token)


def show(kp, args):
    entry = _get(args.path, kp)
    attrs = ['path', 'username', 'password', 'url', 'autotype_sequence',
              'notes']

    for attr in attrs:
        if args.fields and attr not in args.fields:
            continue
        val = getattr(entry, attr)
        if val is None:
            continue

        if args.no_field_name:
            print(val)
        else:
            print('{}: {}'.format(attr, val))


def echo(kp, args):
    print(*args.message)


def sleep(kp, args):
    time.sleep(args.secs)


def ls(kp, args):
    for path in fnmatch.filter(kp.iter_paths(), args.glob):
        print(path)


def help_(kp, args, parsers):
    parser = parsers.get(args.command)
    if parser is None:
        eprint('no such command: {}'.format(args.command))
        parser = parsers.get(None)

    parser.print_help()


def exit(*args):
    return False


def open_(kp, args):
    fp = os.path.expanduser(args.filepath)
    kp.change_db(fp)


def unlock(kp, args):
    if not kp.locked:
        return

    kf = os.path.expanduser(args.keyfile) if args.keyfile else None
    kp.change_credentials(keyfile=kf)
    kp.unlock()


def lock(kp, args):
    kp.lock()


def _get(path, kp):
    entry = kp.entries.get(path)
    if not entry:
        raise CommandError('entry not found: {}'.format(path))
    return entry
